from __future__ import annotations

import gzip
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

import pytest

if TYPE_CHECKING:
    from collections.abc import Iterator


def absolute_path(filename: str) -> Path:
    return Path(__file__).parent / filename


def open_file(name: str, mode: str = "rb") -> Iterator[BinaryIO]:
    with absolute_path(name).open(mode) as f:
        yield f


def open_file_gz(name: str, mode: str = "rb") -> Iterator[BinaryIO]:
    with gzip.GzipFile(absolute_path(name), mode) as f:
        yield f


@pytest.fixture
def basic_wim_4k() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/basic4k.wim.gz")


@pytest.fixture
def basic_wim_8k() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/basic8k.wim.gz")


@pytest.fixture
def basic_wim_16k() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/basic16k.wim.gz")


@pytest.fixture
def basic_wim_32k() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/basic32k.wim.gz")


@pytest.fixture
def basic_vma() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/test.vma.gz")


@pytest.fixture
def vbk9() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/test9.vbk.gz")


@pytest.fixture
def vbk13() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/test13.vbk.gz")



from __future__ import annotations

import hashlib
import struct
from typing import BinaryIO
from unittest.mock import Mock

from dissect.archive.vbk import (
    PAGE_SIZE,
    VBK,
    MetaBlob,
    MetaTableDescriptor,
    MetaVector,
    MetaVector2,
)


def test_vbk_version_9(vbk9: BinaryIO) -> None:
    """test VBK parsing for version 9 files"""
    vbk = VBK(vbk9)

    assert vbk.format_version == 9
    assert vbk.is_v7()
    assert isinstance(vbk.block_store, MetaVector)

    assert vbk.root.is_dir()
    assert not vbk.root.is_file()
    assert list(vbk.get("/").listdir().keys()) == [
        "6745a759-2205-4cd2-b172-8ec8f7e60ef8 (78a5467d-87f5-8540-9a84-7569ae2849ad_2d1bb20f-49c1-485d-a689-696693713a5a)"  # noqa: E501
    ]

    entry = vbk.get(
        "6745a759-2205-4cd2-b172-8ec8f7e60ef8 (78a5467d-87f5-8540-9a84-7569ae2849ad_2d1bb20f-49c1-485d-a689-696693713a5a)"  # noqa: E501
    )
    assert entry.is_dir()
    assert not vbk.root.is_file()
    assert list(entry.listdir().keys()) == [
        "DEV__dev_nvme1n1",
        "summary.xml",
    ]

    entry = vbk.get("DEV__dev_nvme1n1", entry)
    assert not entry.is_dir()
    assert entry.is_file()
    assert entry.is_internal_file()
    assert not entry.properties
    assert entry.size == 0x400000

    with entry.open() as fh:
        digest = hashlib.sha256(fh.read()).hexdigest()
        assert digest == "337350cac29d2ed34c23ce9fc675950badf85fd2b694791abe6999d36f0dc1b3"


def test_vbk_version_13(vbk13: BinaryIO) -> None:
    """test VBK parsing for version 13 files"""
    vbk = VBK(vbk13)

    assert vbk.format_version == 13
    assert vbk.is_v7()
    assert isinstance(vbk.block_store, MetaVector2)
    assert list(vbk.get("/").listdir().keys()) == [
        "6745a759-2205-4cd2-b172-8ec8f7e60ef8 (3c834d56-37ac-8bd3-b946-30113c55c4b5)"
    ]

    entry = vbk.get("6745a759-2205-4cd2-b172-8ec8f7e60ef8 (3c834d56-37ac-8bd3-b946-30113c55c4b5)")
    assert entry.is_dir()
    assert not entry.is_file()
    assert list(entry.listdir().keys()) == [
        "digest_47d9f323-442b-433d-bd4f-1ecb3fa97351",
        "8b14f74c-360d-4d7a-98f7-7f4c5e737eb7",
        "GuestMembers.xml",
        "BackupComponents.xml",
        "summary.xml",
    ]

    entry = vbk.get(
        "6745a759-2205-4cd2-b172-8ec8f7e60ef8 (3c834d56-37ac-8bd3-b946-30113c55c4b5)/8b14f74c-360d-4d7a-98f7-7f4c5e737eb7"  # noqa: E501
    )
    assert not entry.is_dir()
    assert entry.is_file()
    assert entry.is_internal_file()
    assert "DefinedBlocksMask" in entry.properties
    assert len(entry.properties["DefinedBlocksMask"]) == 35
    assert entry.size == 0x314200

    with entry.open() as fh:
        digest = hashlib.sha256(fh.read()).hexdigest()
        assert digest == "e9ed281cf9c2fe1745e4eb9c926c1a64bd47569c48be511c5fdf6fd5793e5a77"


def test_metavector2_lookup() -> None:
    """test that the lookup logic in MetaVector2 works as expected"""

    raw_blob = []
    entry = 0
    num_pages = 11
    for i in range(num_pages):
        if i == 0:
            # root page
            header = struct.pack("<qq", i + 1, 0)
        elif i % 3 == 1:
            # "first" page
            header = struct.pack(
                "<qqqq",
                i + 1 if i + 1 < num_pages else -1,
                i,
                i + 1 if i + 1 < num_pages else -1,
                i + 2 if i + 2 < num_pages else -1,
            )
        else:
            # "middle" pages
            header = struct.pack("<q", i + 1 if i + 1 < num_pages else -1)

        num_entries = (PAGE_SIZE // 8) - (len(header) // 8)
        data = struct.pack(f"<{num_entries}q", *range(entry, entry + num_entries))
        raw_blob.append(header + data)
        entry += num_entries

    mock_blob = b"".join(raw_blob)

    mock_slot = Mock()
    mock_slot.page = lambda page: mock_blob[page * PAGE_SIZE : (page + 1) * PAGE_SIZE]
    mock_vbk = Mock(format_version=13)
    mock_vbk.get_meta_blob = lambda page: MetaBlob(mock_slot, page)
    vec = MetaVector2(mock_vbk, MetaTableDescriptor, 0, 4096)
    assert isinstance(vec, MetaVector2)

    for i in range(entry):
        assert vec._lookup_page(i) == i



from __future__ import annotations

import hashlib
from typing import BinaryIO

from dissect.archive.vma import VMA, _iter_clusters


def test_vma(basic_vma: BinaryIO) -> None:
    vma = VMA(basic_vma)

    assert vma.version == 1
    assert str(vma.uuid) == "04fc12eb-0fed-4322-9aaa-f4e412f68096"

    assert vma.blob_string(1) == "qemu-server.conf"
    assert len(vma.blob_data(20)) == 417
    assert vma.blob_string(439) == "drive-scsi0"

    assert vma.config("qemu-server.conf") == vma.blob_data(20)
    assert len(vma.configs()) == 1

    assert len(vma.devices()) == 1

    device = vma.device(1)
    assert device.id == 1
    assert device.name == "drive-scsi0"
    assert device.size == 10737418240

    extents = list(vma.extents())
    # The test data is just a small piece of a real VMA file
    assert len(extents) == 2

    assert list(_iter_clusters(vma, device.id, 0, 23)) == [
        (0, 65535, 13312),
        (1, 0, 78848),
        (2, 0, 78848),
        (3, 0, 78848),
        (4, 0, 78848),
        (5, 0, 78848),
        (6, 0, 78848),
        (7, 0, 78848),
        (8, 0, 78848),
        (9, 0, 78848),
        (10, 0, 78848),
        (11, 0, 78848),
        (12, 0, 78848),
        (13, 0, 78848),
        (14, 0, 78848),
        (15, 0, 78848),
        (16, 65535, 79360),
        (17, 65535, 144896),
        (18, 65535, 210432),
        (19, 65535, 275968),
        (20, 65535, 341504),
        (21, 65535, 407040),
        (22, 65535, 472576),
    ]

    stream = device.open()
    buf = stream.read(65536)
    assert hashlib.sha256(buf).hexdigest() == "cf4adcf1933a8c9a0a3ff5588e1400e6beea8a32212b3a35ba08c7b08e4e6b1f"

    buf = stream.read(65536 * 15)
    assert buf.strip(b"\x00") == b""

    buf = stream.read(65536 * 7)
    assert hashlib.sha256(buf).hexdigest() == "8c989a3aa590795fa919ccb7d1f28651c85f8a0d9ba00ab22cdd9fb760fa7955"
